"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.jsonschema import Validator
from cfnlint.rules.formats.FormatKeyword import FormatKeyword


class ImageId(FormatKeyword):
    id = "E1152"
    shortdesc = "Validate AMI id format"
    description = "Check that a AMI id matches a pattern"
    tags = []
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/format_keyword.md#AWS::EC2::Image.Id"

    def __init__(self):
        super().__init__(format="AWS::EC2::Image.Id")

    def format(self, validator: Validator, instance: Any) -> bool:
        if not isinstance(instance, str):
            return True

        if validator.context.path.cfn_path_string == (
            "Resources/AWS::EC2::LaunchTemplate/"
            "Properties/LaunchTemplateData/ImageId"
        ):
            if instance.startswith("resolve:ssm"):
                return True

        if re.match(r"^ami-([0-9a-z]{8}|[0-9a-z]{17})$", instance):
            return True

        return False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.SubParametersUsed import SubParametersUsed


@pytest.fixture(scope="module")
def rule():
    rule = SubParametersUsed()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        ("Valid with matching parameters", ["${Foo}", {"Foo": "Bar"}], []),
        (
            "Invalid with a missing parameter",
            ["${Foo}", {"Foo": "Bar", "Bar": "Foo"}],
            [ValidationError("Parameter 'Bar' not used in 'Fn::Sub'", path=[1, "Bar"])],
        ),
    ],
)
def test_sub_parameters_used(name, instance, expected, rule, validator):
    errors = list(rule.validate(validator, {}, instance, {}))
    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import os
from pathlib import Path

from cfnlint._typing import RuleMatches
from cfnlint.helpers import LIMITS
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class LimitSize(CloudFormationLintRule):
    """Check Template Size"""

    id = "E1002"
    shortdesc = "Validate if a template size is too large"
    description = "Check the size of the template is less than the upper limit"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["limits"]

    def match(self, cfn: Template) -> RuleMatches:
        matches = []
        # Only check if the file exists. The template could be passed in using stdIn
        if cfn.filename:
            if Path(cfn.filename).is_file():
                statinfo = os.stat(cfn.filename)
                if statinfo.st_size > LIMITS["template"]["body"]:
                    message = (
                        "The template file size ({0} bytes) exceeds the limit ({1}"
                        " bytes)"
                    )
                    matches.append(
                        RuleMatch(
                            ["Template"],
                            message.format(
                                statinfo.st_size, LIMITS["template"]["body"]
                            ),
                        )
                    )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.elasticache.CacheClusterFailover import (
    CacheClusterFailover,  # pylint: disable=E0401
)


class TestElasticCacheClusterFailover(BaseRuleTestCase):
    """Test ElasticCache CacheClusterFailover"""

    def setUp(self):
        """Setup"""
        super(TestElasticCacheClusterFailover, self).setUp()
        self.collection.register(CacheClusterFailover())
        self.success_templates = [
            "test/fixtures/templates/good/resources/elasticache/cache_cluster_failover.yaml"
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_artifact_failure(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/elasticache/cache_cluster_failover.yaml",
            5,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CloudFormationLintRule


class SnapStartEnabled(CloudFormationLintRule):
    """Check if the SnapStart is enabled for certain java runtimes"""

    id = "I2530"
    shortdesc = "Validate that SnapStart is configured for >= Java11 runtimes"
    description = (
        "SnapStart is a no-cost feature that can increase performance up to 10x. "
        "Enable SnapStart for Java 11 and greater runtimes"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/snapstart.html"
    tags = ["resources", "lambda"]

    def validate(self, runtime):
        if not isinstance(runtime, str):
            return

        if not (runtime.startswith("java")) or runtime in ["java8.al2", "java8"]:
            return

        yield ValidationError(
            f"{runtime!r} runtime should consider using 'SnapStart'",
            path=deque(["SnapStart", "ApplyOn"]),
            rule=self,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.rules.functions._BaseFn import BaseFn


class Condition(BaseFn):
    """Check And Condition Function Logic"""

    id = "E8007"
    shortdesc = "Check Condition structure for validity"
    description = "Check Condition has a value of another condition"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-and"
    tags = ["functions", "and"]

    def __init__(self) -> None:
        super().__init__("Condition", ("boolean",))
        self.condition = self.validate

    def schema(self, validator, instance) -> dict[str, Any]:
        return {
            "type": "string",
            "enum": list(validator.context.conditions.conditions.keys()),
        }



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_elasticloadbalancingv2_loadbalancer
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class LoadBalancerApplicationSubnets(CfnLintJsonSchema):
    id = "E3680"
    shortdesc = "Application load balancers require at least 2 subnets"
    description = ""
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ElasticLoadBalancingV2::LoadBalancer/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_elasticloadbalancingv2_loadbalancer,
                filename="application_subnets.json",
            ),
            all_matches=True,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.apigateway.RestApiOpenApi import RestApiOpenApi


@pytest.fixture(scope="module")
def rule():
    rule = RestApiOpenApi()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Name": "Foo",
            },
            [],
        ),
        (
            {
                "Body": "Foo",
            },
            [],
        ),
        (
            {
                "BodyS3Location": "s3://foo",
            },
            [],
        ),
        (
            {},
            [
                ValidationError(
                    (
                        "'Name' is a required property when not specifying "
                        "one of ['Body', 'BodyS3Location']"
                    ),
                    rule=RestApiOpenApi(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["else", "required"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbinstance
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class DbInstanceEngineVersion(CfnLintJsonSchema):
    id = "E3691"
    shortdesc = "Validate DB Instance Engine and Engine Version"
    description = "Validate the DB Instance engine along with the engine version"
    tags = ["resources"]
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-rds-dbinstance.html#cfn-rds-dbinstance-engine"

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::RDS::DBInstance/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_rds_dbinstance,
                filename="engine_version.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        if validator.is_type(instance.get("Engine"), "string"):
            instance["Engine"] = instance["Engine"].lower()

        yield from super().validate(validator, keywords, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class RoleArnPattern(CfnLintKeyword):
    """Check role arn pattern"""

    id = "E3511"
    shortdesc = "Validate IAM role arn pattern"
    description = "Validate an IAM role arn pattern matches"
    source_url = (
        "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html"
    )
    tags = ["parameters", "iam"]

    def __init__(self) -> None:
        super().__init__(
            [
                "Resources/AWS::Backup::BackupSelection/Properties/BackupSelection/IamRoleArn",
                "Resources/AWS::Batch::ComputeEnvironment/Properties/ComputeResources/SpotIamFleetRole",
                "Resources/AWS::Batch::ComputeEnvironment/Properties/ServiceRole",
                "Resources/AWS::EC2::SpotFleet/Properties/SpotFleetRequestConfigData/IamFleetRole",
                "Resources/AWS::ECS::TaskDefinition/Properties/ExecutionRoleArn",
                "Resources/AWS::S3::Bucket/Properties/ReplicationConfiguration/Role",
            ]
        )

    # pylint: disable=unused-argument
    def validate(
        self, validator: Validator, aZ: Any, arn: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(arn, "string"):
            return

        patrn = "^arn:(aws[a-zA-Z-]*)?:iam::\\d{12}:role/[a-zA-Z_0-9+=,.@\\-_/]+$"

        if not re.match(patrn, arn):
            yield ValidationError(
                f"{arn!r} does not match {patrn!r}",
                rule=self,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

import cfnlint.decode.cfn_json  # pylint: disable=E0401
import cfnlint.decode.cfn_yaml  # pylint: disable=E0401


class TestNonObjectTemplate(BaseTestCase):
    """Test Duplicates Parsing"""

    def setUp(self):
        """SetUp template object"""

    def test_fail_yaml_run(self):
        """Test failure run"""

        filename = "test/fixtures/templates/bad/string.yaml"

        _, matches = cfnlint.decode.decode(filename)
        assert len(matches) == 1



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint._typing import Path, RuleMatches
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class CodepipelineStageActions(CloudFormationLintRule):
    """Check if CodePipeline Stage Actions are set up properly."""

    id = "E2541"
    shortdesc = "CodePipeline Stage Actions"
    description = "See if CodePipeline stage actions are set correctly"
    source_url = "https://docs.aws.amazon.com/codepipeline/latest/userguide/reference-pipeline-structure.html#pipeline-requirements"
    tags = ["resources", "codepipeline"]

    CONSTRAINTS = {
        "AWS": {
            "Source": {
                "S3": {
                    "InputArtifactRange": 0,
                    "OutputArtifactRange": 1,
                },
                "CodeCommit": {
                    "InputArtifactRange": 0,
                    "OutputArtifactRange": 1,
                },
                "ECR": {
                    "InputArtifactRange": 0,
                    "OutputArtifactRange": 1,
                },
            },
            "Test": {
                "CodeBuild": {
                    "InputArtifactRange": (1, 5),
                    "OutputArtifactRange": (0, 5),
                },
                "DeviceFarm": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
            },
            "Build": {
                "CodeBuild": {
                    "InputArtifactRange": (1, 5),
                    "OutputArtifactRange": (0, 5),
                }
            },
            "Approval": {
                "Manual": {
                    "InputArtifactRange": 0,
                    "OutputArtifactRange": 0,
                }
            },
            "Deploy": {
                "S3": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
                "CloudFormation": {
                    "InputArtifactRange": (0, 10),
                    "OutputArtifactRange": (0, 1),
                },
                "CodeDeploy": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
                "ElasticBeanstalk": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
                "OpsWorks": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
                "ECS": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
                "ServiceCatalog": {
                    "InputArtifactRange": 1,
                    "OutputArtifactRange": 0,
                },
            },
            "Invoke": {
                "Lambda": {
                    "InputArtifactRange": (0, 5),
                    "OutputArtifactRange": (0, 5),
                }
            },
        },
        "ThirdParty": {
            "Source": {
                "GitHub": {
                    "InputArtifactRange": 0,
                    "OutputArtifactRange": 1,
                }
            },
            "Deploy": {
                "AlexaSkillsKit": {
                    "InputArtifactRange": (0, 2),
                    "OutputArtifactRange": 0,
                },
            },
        },
        "Custom": {
            "Build": {
                "Jenkins": {
                    "InputArtifactRange": (0, 5),
                    "OutputArtifactRange": (0, 5),
                },
            },
            "Test": {
                "Jenkins": {
                    "InputArtifactRange": (0, 5),
                    "OutputArtifactRange": (0, 5),
                },
            },
        },
    }

    KEY_MAP = {
        "InputArtifacts": "InputArtifactRange",
        "OutputArtifacts": "OutputArtifactRange",
    }

    def check_artifact_counts(self, action, artifact_type, path, scenario):
        """Check that artifact counts are within valid ranges."""
        matches = []

        action_type_id = action.get("ActionTypeId")
        owner = action_type_id.get("Owner")
        category = action_type_id.get("Category")
        provider = action_type_id.get("Provider")

        if (
            isinstance(owner, dict)
            or isinstance(category, dict)
            or isinstance(provider, dict)
        ):
            self.logger.debug(
                "owner, category, provider need to be strings to validate. Skipping."
            )
            return matches

        constraints = (
            self.CONSTRAINTS.get(owner, {}).get(category, {}).get(provider, {})
        )
        if not constraints:
            return matches
        artifact_count = len(action.get(artifact_type, []))

        constraint_key = self.KEY_MAP[artifact_type]
        if isinstance(constraints[constraint_key], tuple):
            min_, max_ = constraints[constraint_key]
            if not min_ <= artifact_count <= max_:
                message = (
                    f'Action "{action["Name"]}" declares'
                    f" {artifact_count} {artifact_type} which is not in expected range"
                    f" [{min_}, {max_}]."
                )
                if scenario:
                    scenario_text = " and ".join(
                        [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    message = message + " When " + scenario_text
                matches.append(RuleMatch(path + [artifact_type], message))
        else:
            if artifact_count != constraints[constraint_key]:
                message = (
                    f'Action "{action["Name"]}" declares'
                    f" {artifact_count} {artifact_type} which is not the expected"
                    f" number [{constraints[constraint_key]}]."
                )
                if scenario:
                    scenario_text = " and ".join(
                        [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    message = message + " When " + scenario_text
                matches.append(RuleMatch(path + [artifact_type], message))

        return matches

    def check_version(self, action, path, scenario):
        """Check that action type version is valid."""
        matches = []

        REGEX_VERSION_STRING = re.compile(r"^[0-9A-Za-z_-]+$")
        LENGTH_MIN = 1
        LENGTH_MAX = 9

        version = action.get("ActionTypeId", {}).get("Version")
        if isinstance(version, dict):
            self.logger.debug(
                "Unable to validate version when an object is used.  Skipping"
            )
        elif isinstance(version, (str)):
            if not LENGTH_MIN <= len(version) <= LENGTH_MAX:
                message = (
                    "Version string ({0}) must be between {1} and {2} characters in"
                    " length."
                )
                if scenario:
                    scenario_text = " and ".join(
                        [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    message = message + " When " + scenario_text
                matches.append(
                    RuleMatch(
                        path + ["ActionTypeId", "Version"],
                        message.format(version, LENGTH_MIN, LENGTH_MAX),
                    )
                )
            elif not re.match(REGEX_VERSION_STRING, version):
                message = "Version string must match the pattern [0-9A-Za-z_-]+."
                if scenario:
                    scenario_text = " and ".join(
                        [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    message = message + " When " + scenario_text
                matches.append(RuleMatch(path + ["ActionTypeId", "Version"], message))
        return matches

    def check_names_unique(self, action, path, action_names, scenario):
        """Check that action names are unique."""
        matches = []

        action_name = action.get("Name")
        if isinstance(action_name, str):
            if action.get("Name") in action_names:
                message = (
                    "All action names within a stage must be unique"
                    f" ({action.get('Name')})."
                )
                if scenario:
                    scenario_text = " and ".join(
                        [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    message = message + " When " + scenario_text
                matches.append(RuleMatch(path + ["Name"], message))
            action_names.add(action.get("Name"))

        return matches

    def check_artifact_names(self, action, path, artifact_names, scenario):
        """
        Check that output artifact names are unique and
        inputs are from previous stage outputs.
        """
        matches = []

        input_artifacts = action.get("InputArtifacts")
        if isinstance(input_artifacts, list):
            for input_artifact in input_artifacts:
                artifact_name = input_artifact.get("Name")
                if isinstance(artifact_name, str):
                    if artifact_name not in artifact_names:
                        message = (
                            "Every input artifact for an action must match the output"
                            " artifact of an action earlier in the pipeline"
                            f" ({artifact_name})."
                        )
                        if scenario:
                            scenario_text = " and ".join(
                                [
                                    f'condition "{k}" is {v}'
                                    for (k, v) in scenario.items()
                                ]
                            )
                            message = message + " When " + scenario_text
                        matches.append(
                            RuleMatch(path + ["InputArtifacts", "Name"], message)
                        )

        output_artifacts = action.get("OutputArtifacts")
        if isinstance(output_artifacts, list):
            for output_artifact in output_artifacts:
                artifact_name = output_artifact.get("Name")
                if isinstance(artifact_name, str):
                    if artifact_name in artifact_names:
                        message = (
                            "Every output artifact in the pipeline must have a unique"
                            f" name. ({artifact_name})"
                        )
                        matches.append(
                            RuleMatch(path + ["OutputArtifacts", "Name"], message)
                        )
                    artifact_names.add(artifact_name)

        return matches

    def match(self, cfn: Template) -> RuleMatches:
        """Check that stage actions are set up properly."""
        matches: RuleMatches = []

        for resource_name, resource_value in cfn.get_resources(
            "AWS::CodePipeline::Pipeline"
        ).items():
            path: Path = ["Resources", resource_name, "Properties"]
            properties = resource_value.get("Properties", {})
            scenarios = cfn.get_object_without_nested_conditions(properties, path)
            for scenario in scenarios:
                conditions = scenario.get("Scenario")
                path = path + ["Stages"]
                properties = scenario.get("Object")
                artifact_names: set[str] = set()

                s_stages = properties.get("Stages")
                if not isinstance(s_stages, list):
                    self.logger.debug(
                        "Stages not list. Should have been caught by generic linting."
                    )
                    return matches
                for s_stage_i, s_stage_v in enumerate(s_stages):
                    action_names: set[str] = set()
                    s_actions = s_stage_v.get("Actions")
                    if not isinstance(s_actions, list):
                        self.logger.debug(
                            "Actions not list. Should have been caught by generic"
                            " linting."
                        )
                        return matches

                    for s_action_i, s_action_v in enumerate(s_actions):
                        try:
                            full_path = path + [s_stage_i, "Actions", s_action_i]
                            matches.extend(
                                self.check_names_unique(
                                    s_action_v, full_path, action_names, conditions
                                )
                            )
                            matches.extend(
                                self.check_version(s_action_v, full_path, conditions)
                            )
                            matches.extend(
                                self.check_artifact_counts(
                                    s_action_v, "InputArtifacts", full_path, conditions
                                )
                            )
                            matches.extend(
                                self.check_artifact_counts(
                                    s_action_v, "OutputArtifacts", full_path, conditions
                                )
                            )
                            matches.extend(
                                self.check_artifact_names(
                                    s_action_v, full_path, artifact_names, conditions
                                )
                            )
                        except AttributeError as err:
                            self.logger.debug(
                                (
                                    "Got AttributeError. Should have been caught by"
                                    " generic linting. Ignoring the error here: %s"
                                ),
                                str(err),
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_managedblockchain_node
from cfnlint.rules.jsonschema.CfnLintJsonSchema import SchemaDetails
from cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional import CfnLintJsonSchemaRegional


class NodeNodeConfigurationInstanceTypeEnum(CfnLintJsonSchemaRegional):
    id = "E3617"
    shortdesc = "Validate ManagedBlockchain instance type"
    description = (
        "Validates the ManagedBlockchain instance types based on region "
        "and data gathered from the pricing APIs"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ManagedBlockchain::Node/Properties/NodeConfiguration/InstanceType"
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_managedblockchain_node,
                filename="nodeconfiguration_instancetype_enum.json",
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint import ConfigMixIn
from cfnlint.rules.resources.HardCodedArnProperties import (
    HardCodedArnProperties,  # pylint: disable=E0401
)


class TestHardCodedArnProperties(BaseRuleTestCase):
    """Test template parameter configurations"""

    def setUp(self):
        """Setup"""
        super(TestHardCodedArnProperties, self).setUp()
        self.collection.register(HardCodedArnProperties())
        self.success_templates = [
            "test/fixtures/templates/good/resources/properties/hard_coded_arn_properties_sam.yaml",
        ]

    def test_file_positive(self):
        """Test Positive"""
        # By default, a set of "correct" templates are checked
        self.helper_file_positive()

    def test_file_negative_partition(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/hard_coded_arn_properties.yaml",
            2,
            ConfigMixIn(
                [],
                include_experimental=True,
                include_checks=[
                    "I",
                ],
                configure_rules={
                    "I3042": {
                        "partition": True,
                        "region": False,
                        "accountId": False,
                    }
                },
            ),
        )

    def test_file_negative_region(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/hard_coded_arn_properties.yaml",
            4,
            ConfigMixIn(
                [],
                include_experimental=True,
                include_checks=[
                    "I",
                ],
                configure_rules={
                    "I3042": {
                        "partition": False,
                        "region": True,
                        "accountId": False,
                    }
                },
            ),
        )

    def test_file_negative_accountid(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/hard_coded_arn_properties.yaml",
            1,
            ConfigMixIn(
                [],
                include_experimental=True,
                include_checks=[
                    "I",
                ],
                configure_rules={
                    "I3042": {
                        "partition": False,
                        "region": False,
                        "accountId": True,
                    }
                },
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.jsonschema.PropertyNames import PropertyNames as ParentPropertyNames


class PropertyNames(ParentPropertyNames):
    id = "E2011"
    shortdesc = "Validate the name for a parameter"
    description = (
        "Validate the name of a parameter with special handling "
        "of the max length length"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["parameters", "limits"]

    # by bundling under propertyNames we can include one rule for all property
    # name validation

    def __init__(self) -> None:
        super().__init__("I2011")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import Validator
from cfnlint.rules.functions._BaseFn import BaseFn, all_types


class Select(BaseFn):
    """Check if Select values are correct"""

    id = "E1017"
    shortdesc = "Select validation of parameters"
    description = "Making sure the Select function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-select.html"
    tags = ["functions", "select"]

    def __init__(self) -> None:
        super().__init__("Fn::Select", all_types, resolved_rule="W1035")
        self.fn_select = self.validate

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": "array",
            "maxItems": 2,
            "minItems": 2,
            "fn_items": [
                {
                    "functions": ["Ref", "Fn::FindInMap"],
                    "schema": {
                        "type": ["integer"],
                    },
                },
                {
                    "functions": [
                        "Fn::FindInMap",
                        "Fn::GetAtt",
                        "Fn::GetAZs",
                        "Fn::If",
                        "Fn::Split",
                        "Fn::Cidr",
                        "Ref",
                    ],
                    "schema": {
                        "type": ["array"],
                        "fn_items": {
                            "functions": [
                                "Fn::FindInMap",
                                "Fn::GetAtt",
                                "Fn::If",
                                "Ref",
                            ],
                            "schema": {
                                "type": all_types,
                            },
                        },
                    },
                },
            ],
        }



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase
from unittest.mock import MagicMock, patch

import cfnlint.helpers


class TestGetUrlRetrieve(BaseTestCase):
    """Test Get URL Retrieve"""

    @patch("cfnlint.helpers.urlretrieve")
    def test_get_url_retrieve(self, mocked_urlretrieve):
        """Test the basics of URL retrieve"""

        mocked_urlretrieve.return_value = ("file/path", None)

        url = "http://foo.com"
        result = cfnlint.helpers.get_url_retrieve(url)
        mocked_urlretrieve.assert_called_with(url)
        self.assertEqual(result, "file/path")

    @patch("cfnlint.helpers.urlopen")
    @patch("cfnlint.helpers.load_metadata")
    @patch("cfnlint.helpers.save_metadata")
    @patch("cfnlint.helpers.urlretrieve")
    def test_get_url_retrieve_cached(
        self, mocked_urlretrieve, mock_save_metadata, mock_load_metadata, mocked_urlopen
    ):
        """Test the basics of URL retrieve"""
        etag = "ETAG_ONE"
        url = "http://foo.com"

        mock_load_metadata.return_value = {}

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.info.return_value = {"Content-Encoding": "gzip", "ETag": etag}
        cm.__enter__.return_value = cm
        mocked_urlopen.return_value = cm

        mocked_urlretrieve.return_value = ("file/path", None)

        result = cfnlint.helpers.get_url_retrieve(url, caching=True)
        mocked_urlretrieve.assert_called_with(url)
        mock_load_metadata.assert_called_once()
        mock_save_metadata.assert_called_once()
        self.assertEqual(result, "file/path")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class Equals(BaseFn):
    """Check Equals Condition Function Logic"""

    id = "E8003"
    shortdesc = "Check Fn::Equals structure for validity"
    description = "Check Fn::Equals is a list of two elements"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-equals"
    tags = ["functions", "equals"]

    def __init__(self) -> None:
        super().__init__("Fn::Equals", ("boolean",))
        self.child_rules = {
            "W8003": None,
        }

    def fn_equals(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        errs = list(self.validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)
            return

        child_rule = self.child_rules.get("W8003")
        if child_rule and hasattr(child_rule, "equals_is_useful"):
            yield from child_rule.equals_is_useful(
                validator, s, instance.get("Fn::Equals", []), schema
            )

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": "array",
            "maxItems": 2,
            "minItems": 2,
            "fn_items": {
                "functions": [
                    "Ref",
                    "Fn::FindInMap",
                    "Fn::Sub",
                    "Fn::Join",
                    "Fn::Select",
                    "Fn::Split",
                    "Fn::Length",
                    "Fn::ToJsonString",
                ],
                "schema": {
                    "type": ["string"],
                },
            },
        }



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.elb.Certificate import Certificate


@pytest.fixture(scope="module")
def rule():
    rule = Certificate()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"Protocol": "HTTP"},
            [],
        ),
        (
            {"Protocol": "HTTPS", "SSLCertificateId": "Id"},
            [],
        ),
        (
            {
                "Protocol": "SSL",
            },
            [
                ValidationError(
                    ("'SSLCertificateId' is a required property"),
                    rule=Certificate(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "required"]),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class NumberRange(CloudFormationLintRule):
    """Check if a Number has a length within the limit"""

    id = "W3034"
    shortdesc = "Check if parameter values are between min and max"
    description = (
        "Check if parameter values value being between the minimum" " and maximum"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#number-size"
    tags = ["resources", "property", "number", "size"]

    def validate(self, validator, m, instance, schema, fn):
        for err in fn(validator, m, instance, schema):
            err.rule = self
            err.path_override = validator.context.path.value_path
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import importlib
import logging
import os
import traceback
from collections import UserDict
from typing import TYPE_CHECKING, Any, Callable, Iterator, MutableSet

import cfnlint.helpers
import cfnlint.rules.custom
from cfnlint.exceptions import DuplicateRuleError
from cfnlint.rules._rule import CloudFormationLintRule, Match
from cfnlint.rules.errors import RuleError
from cfnlint.template import Template

if TYPE_CHECKING:
    from cfnlint.config import ConfigMixIn

    TypedRules = UserDict[str, "CloudFormationLintRule"]
else:
    TypedRules = UserDict


LOGGER = logging.getLogger(__name__)


class Rules(TypedRules):
    def __init__(
        self, rules: dict[str, CloudFormationLintRule] | None = None, /, **kwargs
    ):
        super().__init__()
        self.data: dict[str, CloudFormationLintRule] = {}
        self._used_rules: dict[str, CloudFormationLintRule] = {}
        if rules is not None:
            self.update(rules)
        if kwargs:
            self.update(kwargs)

    def __repr__(self):
        return "\n".join([self.data[id].verbose() for id in sorted(self.data)])

    def __delitem__(self, i: str) -> None:
        raise RuntimeError("Deletion is not allowed")

    def __setitem__(self, key: str, item: CloudFormationLintRule) -> None:
        if not key:
            return
        if key in self.data:
            raise DuplicateRuleError(rule_id=key)
        return super().__setitem__(key, item)

    def register(self, rule: CloudFormationLintRule) -> None:
        self[rule.id] = rule

    def filter(
        self,
        func: Callable[[dict[str, CloudFormationLintRule], ConfigMixIn], Rules],
        config: ConfigMixIn,
    ):
        return func(self.data, config)

    def is_rule_enabled(
        self, rule: str | CloudFormationLintRule, config: ConfigMixIn
    ) -> bool:
        if isinstance(rule, str):
            if rule not in self.data:
                return False
            rule = self.data[rule]
        if rule.is_enabled(
            include_experimental=config.include_experimental,
            ignore_rules=config.ignore_checks,
            include_rules=config.include_checks,
            mandatory_rules=config.mandatory_checks,
        ):
            return True
        return False

    def extend(self, rules: list[CloudFormationLintRule]):
        for rule in rules:
            self.register(rule)

    @property
    def used_rules(self) -> dict[str, CloudFormationLintRule]:
        return self._used_rules

    # pylint: disable=inconsistent-return-statements
    def run_check(self, check, filename, rule_id, config, *args) -> Iterator[Match]:
        """Run a check"""
        if self.is_rule_enabled(rule_id, config):
            self._used_rules[rule_id] = self.data[rule_id]
        try:
            yield from iter(check(*args))
        except Exception as err:  # pylint: disable=W0703
            if self.is_rule_enabled(RuleError(), config):
                # In debug mode, print the error include complete stack trace
                if LOGGER.getEffectiveLevel() == logging.DEBUG:
                    error_message = traceback.format_exc()
                else:
                    error_message = str(err)
                yield Match.create(
                    filename=filename,
                    rule=RuleError(),
                    message=(
                        "Unknown exception while processing "
                        f"rule {rule_id}: {error_message!r}"
                    ),
                )

    def _filter_matches(
        self, config: ConfigMixIn, matches: Iterator[Match]
    ) -> Iterator[Match]:
        """Filter matches by config"""
        for match in matches:
            if self.is_rule_enabled(match.rule, config):
                yield match

    def run(
        self, filename: str | None, cfn: Template, config: ConfigMixIn
    ) -> Iterator[Match]:
        """Run rules"""
        for rule_id, rule in self.data.items():
            rule.configure(
                config.configure_rules.get(rule_id, None), config.include_experimental
            )
            rule.initialize(cfn)

        for rule_id, rule in self.data.items():
            for key in rule.child_rules.keys():
                if not any(key == r for r in self.data.keys()):
                    continue
                rule.child_rules[key] = self.data.get(key)
            for parent_rule in rule.parent_rules:
                if parent_rule in self.data:
                    self.data[parent_rule].child_rules[rule_id] = rule

        for rule_id, rule in self.data.items():
            yield from self._filter_matches(
                config,
                self.run_check(rule.matchall, filename, rule_id, config, filename, cfn),
            )

        for resource_name, resource_attributes in cfn.get_resources().items():
            resource_type = resource_attributes.get("Type")
            resource_properties = resource_attributes.get("Properties")
            if isinstance(resource_type, str) and isinstance(resource_properties, dict):
                path = ["Resources", resource_name, "Properties"]
                for rule_id, rule in self.data.items():
                    yield from self._filter_matches(
                        config,
                        self.run_check(
                            rule.matchall_resource_properties,
                            filename,
                            rule_id,
                            config,
                            filename,
                            cfn,
                            resource_properties,
                            resource_type,
                            path,
                        ),
                    )

    @classmethod
    def _from_list(cls, items: list[CloudFormationLintRule]) -> Rules:
        rules = Rules()
        for item in items:
            rules[item.id] = item

        return rules

    @classmethod
    def create_from_module(cls, modpath: str) -> Rules:
        """Create rules from a module import path"""
        mod = importlib.import_module(modpath)
        return cls._from_list(cfnlint.helpers.create_rules(mod))

    @classmethod
    def create_from_directory(cls, rulesdir: str) -> Rules:
        if rulesdir != "":
            return cls._from_list(
                cfnlint.helpers.load_plugins(os.path.expanduser(rulesdir))
            )

        return cls({})

    @classmethod
    def create_from_custom_rules_file(cls, custom_rules_file) -> Rules:
        """Create rules from custom rules file"""
        custom_rules = cls({})
        if custom_rules_file:
            with open(custom_rules_file, encoding="utf-8") as customRules:
                line_number = 1
                for line in customRules:
                    LOGGER.debug("Processing Custom Rule Line %d", line_number)
                    custom_rule = cfnlint.rules.custom.make_rule(line, line_number)
                    if custom_rule:
                        custom_rules[custom_rule.id] = custom_rule
                    line_number += 1

        return custom_rules


# pylint: disable=too-many-instance-attributes
class RulesCollection:
    """Collection of rules"""

    def __init__(
        self,
        ignore_rules: list[str] | None = None,
        include_rules: list[str] | None = None,
        configure_rules: Any = None,
        include_experimental: bool = False,
        mandatory_rules: list[str] | None = None,
    ):
        self.rules: dict[str, CloudFormationLintRule] = {}
        self.all_rules: dict[str, CloudFormationLintRule] = {}
        self.used_rules: MutableSet[str] = set()

        self.configure(
            ignore_rules=ignore_rules,
            include_rules=include_rules,
            configure_rules=configure_rules,
            include_experimental=include_experimental,
            mandatory_rules=mandatory_rules,
        )

    def configure(
        self,
        ignore_rules: list[str] | None = None,
        include_rules: list[str] | None = None,
        configure_rules: Any = None,
        include_experimental: bool = False,
        mandatory_rules: list[str] | None = None,
    ):
        self.rules = {}
        # Whether "experimental" rules should be added
        self.include_experimental = include_experimental

        # Make Ignore Rules not required
        self.ignore_rules = ignore_rules or []
        self.include_rules = include_rules or []
        self.mandatory_rules = mandatory_rules or []
        self.configure_rules = configure_rules or {}
        # by default include 'W' and 'E'
        # 'I' has to be included manually for backwards compabitility
        # Have to add W, E here because integrations don't use config
        for default_rule in ["W", "E"]:
            if default_rule not in self.include_rules:
                self.include_rules.extend([default_rule])

        for rule in self.all_rules.values():
            self.__register(rule)

    def __register(self, rule: CloudFormationLintRule):
        """Register and configure the rule"""
        if self.is_rule_enabled(rule) or rule.child_rules:
            self.used_rules.add(rule.id)
            self.rules[rule.id] = rule
            rule.configure(
                self.configure_rules.get(rule.id, None), self.include_experimental
            )

    def register(self, rule: CloudFormationLintRule):
        """Register rules"""
        # Some rules are inheritited to limit code re-use.
        # These rules have no rule ID so we filter this out
        if rule.id != "":
            if rule.id in self.all_rules:
                raise DuplicateRuleError(rule_id=rule.id)
            self.all_rules[rule.id] = rule
            self.__register(rule)

    def __iter__(self):
        return iter(self.rules.values())

    def __len__(self):
        return len(self.rules.keys())

    def extend(self, more):
        """Extend rules"""
        for rule in more:
            self.register(rule)

    def __repr__(self):
        return "\n".join([self.rules[id].verbose() for id in sorted(self.rules)])

    def is_rule_enabled(self, rule: CloudFormationLintRule):
        """Checks if an individual rule is valid"""
        return rule.is_enabled(
            self.include_experimental,
            self.ignore_rules,
            self.include_rules,
            self.mandatory_rules,
        )

    # pylint: disable=inconsistent-return-statements
    def run_check(self, check, filename, rule_id, *args):
        """Run a check"""
        try:
            matches = []
            for match in check(*args):
                if self.is_rule_enabled(match.rule):
                    matches.append(match)
            return matches
        except Exception as err:  # pylint: disable=W0703
            if self.is_rule_enabled(RuleError()):
                # In debug mode, print the error include complete stack trace
                if LOGGER.getEffectiveLevel() == logging.DEBUG:
                    error_message = traceback.format_exc()
                else:
                    error_message = str(err)
                message = "Unknown exception while processing rule {}: {}"
                return [
                    Match.create(
                        filename=filename,
                        rule=RuleError(),
                        message=message.format(rule_id, error_message),
                    )
                ]

        return []

    def run(self, filename: str | None, cfn: Template, config=None):
        """Run rules"""
        matches = []
        for rule in self.rules.values():
            rule.initialize(cfn)

        for rule in self.rules.values():
            for key in rule.child_rules.keys():
                rule.child_rules[key] = self.rules.get(key)
            for parent_rule in rule.parent_rules:
                if parent_rule in self.rules:
                    self.rules[parent_rule].child_rules[rule.id] = rule

        for rule in self.rules.values():
            matches.extend(
                self.run_check(rule.matchall, filename, rule.id, filename, cfn)
            )

        for resource_name, resource_attributes in cfn.get_resources().items():
            resource_type = resource_attributes.get("Type")
            resource_properties = resource_attributes.get("Properties")
            if isinstance(resource_type, str) and isinstance(resource_properties, dict):
                path = ["Resources", resource_name, "Properties"]
                for rule in self.rules.values():
                    matches.extend(
                        self.run_check(
                            rule.matchall_resource_properties,
                            filename,
                            rule.id,
                            filename,
                            cfn,
                            resource_properties,
                            resource_type,
                            path,
                        )
                    )

        return matches

    def create_from_module(self, modpath):
        """Create rules from a module import path"""
        mod = importlib.import_module(modpath)
        self.extend(cfnlint.helpers.create_rules(mod))

    def create_from_directory(self, rulesdir):
        """Create rules from directory"""
        result = []
        if rulesdir != "":
            result = cfnlint.helpers.load_plugins(os.path.expanduser(rulesdir))
        self.extend(result)

    def create_from_custom_rules_file(self, custom_rules_file):
        """Create rules from custom rules file"""
        custom_rules = []
        if custom_rules_file:
            with open(custom_rules_file, encoding="utf-8") as customRules:
                line_number = 1
                for line in customRules:
                    LOGGER.debug("Processing Custom Rule Line %d", line_number)
                    custom_rule = cfnlint.rules.custom.make_rule(line, line_number)
                    if custom_rule:
                        custom_rules.append(custom_rule)
                    line_number += 1

        self.extend(custom_rules)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.jsonschema.Base import BaseJsonSchema
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails
from cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional import CfnLintJsonSchemaRegional
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword
from cfnlint.rules.jsonschema.MaxProperties import MaxProperties
from cfnlint.rules.jsonschema.PropertyNames import PropertyNames

__all__ = [
    "BaseJsonSchema",
    "CfnLintKeyword",
    "CfnLintJsonSchema",
    "SchemaDetails",
    "CfnLintJsonSchemaRegional",
    "MaxProperties",
    "PropertyNames",
]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging

from cfnlint.api import lint, lint_all
from cfnlint.config import ConfigMixIn, ManualArgs
from cfnlint.rules import Rules
from cfnlint.template import Template
from cfnlint.template.transforms import Transform

LOGGER = logging.getLogger(__name__)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

# pylint: disable=cyclic-import
import cfnlint.rules
from cfnlint._typing import RuleMatches

OPERATOR = [
    "EQUALS",
    "NOT_EQUALS",
    "REGEX_MATCH",
    "==",
    "!=",
    "IN",
    "NOT_IN",
    ">",
    "<",
    ">=",
    "<=",
    "IS DEFINED",
    "IS NOT_DEFINED",
]


def CreateCustomRule(
    rule_id, resourceType, prop, value, error_message, description, shortdesc, rule_func
):
    class CustomRule(cfnlint.rules.CloudFormationLintRule):
        def __init__(
            self,
            rule_id,
            resourceType,
            prop,
            value,
            error_message,
            description,
            shortdesc,
            rule_func,
        ):
            super().__init__()
            self.resource_property_types.append(resourceType)
            self.id = rule_id
            self.property_chain = prop.split(".")
            self.property_value = value
            self.error_message = error_message
            self.description = description
            self.shortdesc = shortdesc
            self.rule_func = rule_func

        def _remaining_inset_properties(self, property_chain):
            if len(property_chain) > 1:
                return property_chain[1:]

            return []

        def _check_value(self, value, path, property_chain, cfn):
            matches = []
            if property_chain:
                new_property_chain = self._remaining_inset_properties(property_chain)
                matches.extend(
                    cfn.check_value(
                        value,
                        property_chain[0],
                        path,
                        check_value=self._check_value,
                        property_chain=new_property_chain,
                        cfn=cfn,
                    )
                )
                return matches
            if value is not None:
                matches.extend(self.rule_func(value, self.property_value, path))
            return matches

        def match_resource_properties(self, properties, _, path, cfn):
            new_property_chain = self._remaining_inset_properties(self.property_chain)
            return cfn.check_value(
                properties,
                self.property_chain[0],
                path,
                check_value=self._check_value,
                property_chain=new_property_chain,
                cfn=cfn,
            )

    return CustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        description,
        shortdesc,
        rule_func,
    )


def CreateCustomIsDefinedRule(rule_id, resourceType, prop, value, error_message):
    class CustomIsDefinedRule(cfnlint.rules.CloudFormationLintRule):
        def __init__(
            self,
            rule_id,
            resourceType,
            prop,
            value,
            error_message,
            description,
            shortdesc,
        ):
            super().__init__()
            self.id = rule_id
            self.resource_property_types.append(resourceType)
            self.property_chain = prop.split(".")
            if value == "DEFINED":
                self.is_defined = True
            elif value == "NOT_DEFINED":
                self.is_defined = False
            else:
                raise ValueError("IS must follow either DEFINED or NOT_DEFINED")
            self.error_message = error_message
            self.description = description
            self.shortdesc = shortdesc

        def _split_inset_properties(self, property_chain):
            if property_chain:
                if len(property_chain) > 1:
                    return property_chain[0], property_chain[1:]
                return property_chain[0], []

            return None, []

        def _check_value_defined(self, value, path, property_chain, cfn, **_):
            matches = []
            child_property, new_property_chain = self._split_inset_properties(
                property_chain
            )
            # Specific for !Ref AWS::NoValue, no callback even for pass_if_null=True
            if len(property_chain) == 1 and value == {
                child_property: {"Ref": "AWS::NoValue"}
            }:
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or f"{path} must be defined"
                    )
                )
                return matches
            if value is None or (
                isinstance(value, dict) and value.get("Ref", None) == "AWS::NoValue"
            ):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or f"{path} must be defined"
                    )
                )
                return matches

            if child_property is not None:
                matches.extend(
                    cfn.check_value(
                        value,
                        child_property,
                        path,
                        check_value=self._check_value_defined,
                        check_ref=self._check_value_defined,
                        check_get_att=self._check_value_defined,
                        check_find_in_map=self._check_value_defined,
                        check_split=self._check_value_defined,
                        check_join=self._check_value_defined,
                        check_import_value=self._check_value_defined,
                        check_sub=self._check_value_defined,
                        pass_if_null=True,
                        property_chain=new_property_chain,
                        cfn=cfn,
                    )
                )
            return matches

        def _check_value_not_defined(self, value, path, property_chain, cfn, **_):
            matches = []
            if value is None:
                return matches
            if len(property_chain) == 0 and value is not None:
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or f"{path} must not be defined"
                    )
                )
                return matches

            child_property, new_property_chain = self._split_inset_properties(
                property_chain
            )
            matches.extend(
                cfn.check_value(
                    value,
                    child_property,
                    path,
                    check_value=self._check_value_not_defined,
                    check_ref=self._check_value_not_defined,
                    check_get_att=self._check_value_not_defined,
                    check_find_in_map=self._check_value_not_defined,
                    check_split=self._check_value_not_defined,
                    check_join=self._check_value_not_defined,
                    check_import_value=self._check_value_not_defined,
                    check_sub=self._check_value_not_defined,
                    property_chain=new_property_chain,
                    cfn=cfn,
                    pass_if_null=True,
                )
            )
            return matches

        def match_resource_properties(self, properties, _, path, cfn):
            child_property, new_property_chain = self._split_inset_properties(
                self.property_chain
            )
            check_fn = (
                self._check_value_defined
                if self.is_defined
                else self._check_value_not_defined
            )
            matches = []
            # here does nothing when the value is not defined,
            # this is checked separately below
            matches.extend(
                cfn.check_value(
                    properties,
                    child_property,
                    path,
                    check_value=check_fn,
                    check_ref=check_fn,
                    check_get_att=check_fn,
                    check_find_in_map=check_fn,
                    check_split=check_fn,
                    check_join=check_fn,
                    check_import_value=check_fn,
                    check_sub=check_fn,
                    property_chain=new_property_chain,
                    cfn=cfn,
                    pass_if_null=True,
                )
            )
            return matches

    return CustomIsDefinedRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=f"Custom rule to check for value is {value}",
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            f" property value is {value}"
        ),
    )


def CreateEqualsRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if str(value).strip().lower() != str(expected_value).strip().lower():
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Must equal check failed"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for equal values",
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value is equal to the specified value."
        ),
        rule_func=rule_func,
    )


def CreateNotEqualsRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if str(value).strip().lower() == str(expected_values).strip().lower():
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Must not equal check failed"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for not equal values",
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value is NOT equal to the specified value."
        ),
        rule_func=rule_func,
    )


def CreateRegexMatchRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if not re.match(expected_values.strip(), str(value).strip()):
            matches.append(
                cfnlint.rules.RuleMatch(path, error_message or "Regex does not match")
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for regex match",
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value match the provided regex pattern."
        ),
        rule_func=rule_func,
    )


def CreateGreaterEqualRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if checkInt(str(value).strip()) and checkInt(str(expected_value).strip()):
            if int(str(value).strip()) < int(str(expected_value).strip()):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or "Greater than check failed"
                    )
                )
        else:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Given values are not numeric"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=(
            "Custom rule to check for if a value is greater than the specified value"
        ),
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value is greater than the specified value."
        ),
        rule_func=rule_func,
    )


def CreateGreaterRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if checkInt(str(value).strip()) and checkInt(str(expected_value).strip()):
            if int(str(value).strip()) <= int(str(expected_value).strip()):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or "Greater than check failed"
                    )
                )
        else:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Given values are not numeric"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=(
            "Custom rule to check for if a value is greater than the specified value"
        ),
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value is greater than the specified value."
        ),
        rule_func=rule_func,
    )


def CreateLesserRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if checkInt(str(value).strip()) and checkInt(str(expected_value).strip()):
            if int(str(value).strip()) >= int(str(expected_value).strip()):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or "Lesser than check failed"
                    )
                )
        else:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Given values are not numeric"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=(
            "Custom rule to check for if a value is lesser than the specified value"
        ),
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value is lesser than the specified value."
        ),
        rule_func=rule_func,
    )


def CreateLesserEqualRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if checkInt(str(value).strip()) and checkInt(str(expected_value).strip()):
            if int(str(value).strip()) > int(str(expected_value).strip()):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or "Lesser than check failed"
                    )
                )
        else:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Given values are not numeric"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=(
            "Custom rule to check for if a value is lesser than the specified value"
        ),
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value is lesser than the specified value."
        ),
        rule_func=rule_func,
    )


def CreateInSetRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if value not in expected_values:
            matches.append(
                cfnlint.rules.RuleMatch(path, error_message or "In set check failed")
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=(
            "Custom rule to check for if a value exists in a list of specified values"
        ),
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value exists inside a list of specified values."
        ),
        rule_func=rule_func,
    )


def CreateNotInSetRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if value in expected_values:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Not in set check failed"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc=(
            "Custom rule to check for if a value does not exist in a list of specified"
            " values"
        ),
        description=(
            "Created from the custom rules parameter. This rule will check if a"
            " property value does not exist inside a list of specified values."
        ),
        rule_func=rule_func,
    )


def CreateInvalidRule(rule_id, operator):
    class InvalidRule(cfnlint.rules.CloudFormationLintRule):
        def __init__(self, rule_id, operator):
            super().__init__()
            self.id = rule_id
            self.operator = operator
            self.description = (
                "Created from the custom rule parameter. This rule is the result of an"
                " invalid configuration of a custom rule."
            )
            self.shortdesc = "Invalid custom rule configuration"

        def match(self, _) -> RuleMatches:
            message = '"{0}" not in supported operators: [{1}]'
            return [
                cfnlint.rules.RuleMatch(
                    [], message.format(str(self.operator), ", ".join(OPERATOR))
                )
            ]

    return InvalidRule(rule_id, operator)


def checkInt(i):
    """Python 2.7 Compatibility - There is no isnumeric() method"""
    try:
        int(i)
        return True
    except ValueError:
        return False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.resources.properties.UniqueItems import UniqueItems
from cfnlint.rules.resources.properties.UniqueItemsAllowed import UniqueItemsAllowed


@pytest.fixture(scope="module")
def rule():
    rule = UniqueItems()
    rule.child_rules["I3037"] = UniqueItemsAllowed()
    yield rule


@pytest.mark.parametrize(
    "instance,uI,schema,expected",
    [
        (["foo", "bar"], True, {}, None),
        (["foo", "bar"], False, {}, None),
        (["foo", "foo"], False, {}, "I3037"),
        (["foo", "foo"], True, {}, "E3037"),
    ],
)
def test_unique_items_fails(instance, uI, expected, rule, schema, validator):
    errs = list(rule.uniqueItems(validator, uI, instance, schema))
    if expected is None:
        assert not errs
    else:
        assert len(errs) == 1
        if expected != "E3037":
            assert errs[0].rule.id == expected



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import VALID_PARAMETER_TYPES, VALID_PARAMETER_TYPES_LIST
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, all_types


class Ref(BaseFn):
    id = "E1020"
    shortdesc = "Ref validation of value"
    description = (
        "Making sure the Ref has a String value (no other functions are supported)"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html"
    tags = ["functions", "ref"]

    def __init__(self) -> None:
        super().__init__("Ref", all_types, resolved_rule="W1030")
        self._all_refs = [
            "W2010",
        ]
        self.child_rules.update(dict.fromkeys(self._all_refs))

    def schema(self, validator, instance) -> dict[str, Any]:
        return {
            "type": ["string"],
            "enum": validator.context.refs,
        }

    def validator(self, validator: Validator) -> Validator:
        if validator.context.transforms.has_language_extensions_transform():
            supported_functions = [
                "Ref",
                "Fn::Base64",
                "Fn::FindInMap",
                "Fn::If",
                "Fn::Join",
                "Fn::Sub",
                "Fn::ToJsonString",
            ]
        else:
            supported_functions = []
        return validator.evolve(
            context=validator.context.evolve(
                functions=supported_functions,
            ),
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=False,
            ),
        )

    def ref(self, validator, subschema, instance, schema):
        yield from super().validate(validator, subschema, instance, schema)

        _, value = self.key_value(instance)
        if not validator.is_type(value, "string"):
            return

        if value not in validator.context.parameters:
            return

        parameter_type = validator.context.parameters[value].type
        schema_types = self.resolve_type(validator, subschema)
        if not schema_types:
            return
        reprs = ", ".join(repr(type) for type in schema_types)

        if all(
            st not in ["string", "boolean", "integer", "number"] for st in schema_types
        ):
            if parameter_type not in VALID_PARAMETER_TYPES_LIST:
                yield ValidationError(f"{instance!r} is not of type {reprs}")
                return
        elif all(st not in ["array"] for st in schema_types):
            if parameter_type not in [
                x for x in VALID_PARAMETER_TYPES if x not in VALID_PARAMETER_TYPES_LIST
            ]:
                yield ValidationError(f"{instance!r} is not of type {reprs}")
                return

        for rule_id in self._all_refs:
            rule = self.child_rules.get(rule_id)
            if rule:
                yield from rule.validate(validator, {}, instance, schema)

        keyword = validator.context.path.cfn_path_string
        for rule in self.child_rules.values():
            if not rule or rule.id in self._all_refs:
                continue
            if not hasattr(rule, "keywords"):
                continue
            if keyword in rule.keywords:
                yield from rule.validate(validator, keyword, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import List

from cfnlint.formatters.base import BaseFormatter
from cfnlint.match import Match
from cfnlint.rules import Rules
from cfnlint.version import __version__

Matches = List[Match]


class SARIFFormatter(BaseFormatter):
    """
    SARIF formatter

    This formatter outputs results according to the Static Analysis Results
    Interchange Format (SARIF) Version 2.1.0.

    https://docs.oasis-open.org/sarif/sarif/v2.1.0/csprd01/sarif-v2.1.0-csprd01.html
    """

    schema = "https://docs.oasis-open.org/sarif/sarif/v2.1.0/cos02/schemas/sarif-schema-2.1.0.json"
    version = "2.1.0"

    # The spec defines error, note, warning, and none, see section 3.27.10.
    levelMap = {
        "error": "error",
        "informational": "note",
        "warning": "warning",
    }

    uri_base_id = "EXECUTIONROOT"

    def _to_sarif_level(self, severity):
        return self.levelMap.get(severity, "none")

    def print_matches(self, matches, rules, config):
        """Output all the matches"""

        try:
            import sarif_om as sarif
            from jschema_to_python.to_json import to_json
        except ImportError as e:
            raise ImportError("Missing optional dependencies sarif") from e

        if not rules:
            rules = Rules()

        results = []
        for match in matches:
            results.append(
                sarif.Result(
                    rule_id=match.rule.id,
                    message=sarif.Message(text=match.message),
                    level=self._to_sarif_level(match.rule.severity),
                    locations=[
                        sarif.Location(
                            physical_location=sarif.PhysicalLocation(
                                artifact_location=sarif.ArtifactLocation(
                                    uri=match.filename,
                                    uri_base_id=self.uri_base_id,
                                ),
                                region=sarif.Region(
                                    start_column=match.columnnumber,
                                    start_line=match.linenumber,
                                    end_column=match.columnnumberend,
                                    end_line=match.linenumberend,
                                ),
                            )
                        )
                    ],
                )
            )

        # Output only the rules that have matches
        matched_rules = set(r.rule_id for r in results)

        rules = [
            sarif.ReportingDescriptor(
                id=rule_id,
                short_description=sarif.MultiformatMessageString(
                    text=rules[rule_id].shortdesc
                ),
                full_description=sarif.MultiformatMessageString(
                    text=rules[rule_id].description
                ),
                help_uri=(
                    rules[rule_id].source_url
                    if rules[rule_id]
                    else "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/rules.md"
                ),
            )
            for rule_id in matched_rules
        ]

        run = sarif.Run(
            tool=sarif.Tool(
                driver=sarif.ToolComponent(
                    name="cfn-lint",
                    short_description=sarif.MultiformatMessageString(
                        text=(
                            "Validates AWS CloudFormation templates against"
                            " the resource specification and additional"
                            " checks."
                        )
                    ),
                    information_uri="https://github.com/aws-cloudformation/cfn-lint",
                    rules=rules,
                    version=__version__,
                ),
            ),
            original_uri_base_ids={
                self.uri_base_id: sarif.ArtifactLocation(
                    description=sarif.MultiformatMessageString(
                        "The directory in which cfn-lint was run."
                    )
                )
            },
            results=results,
        )

        log = sarif.SarifLog(version=self.version, schema_uri=self.schema, runs=[run])

        # IMPORTANT: 'warning' is the default level in SARIF and will be
        # stripped by serialization.
        return to_json(log)



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

from typing import Any

import regex as re

import cfnlint.jsonschema._keywords as validators_standard
from cfnlint.helpers import BOOLEAN_STRINGS, ensure_list
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.jsonschema._typing import V, ValidationResult


def additionalProperties(
    validator: Validator, aP: Any, instance: Any, schema: Any
) -> ValidationResult:
    for err in validators_standard.additionalProperties(
        validator, aP, instance, schema
    ):
        if len(err.path) > 0:
            if any(
                re.fullmatch(fn, str(err.path[0])) for fn in validator.context.functions
            ):
                continue
            yield err
        else:
            yield err


class FnItems:
    def validate(
        self,
        validator: Validator,
        s: Any,
        instance: Any,
        schema: Any,
    ) -> ValidationResult:
        if not validator.is_type(instance, "array"):
            return

        if validator.is_type(s, "array"):
            for (index, item), subschema in zip(enumerate(instance), s):
                yield from validator.evolve(
                    context=validator.context.evolve(
                        functions=subschema.get("functions", []),
                        strict_types=False,
                    ),
                ).descend(
                    instance=item,
                    schema=subschema.get("schema", {}),
                    path=index,
                )
        else:
            for index, item in enumerate(instance):
                yield from validator.evolve(
                    context=validator.context.evolve(
                        functions=s.get("functions", []),
                        strict_types=False,
                    ),
                ).descend(
                    instance=item,
                    schema=s.get("schema", {}),
                    path=index,
                )


#####
# Type checks
#####
def _raw_type(validator: Validator, tS: Any, instance: Any) -> bool:
    if tS in ["object", "array"]:
        return validator.is_type(instance, tS)
    if "string" == tS:
        if validator.is_type(instance, "object") or validator.is_type(
            instance, "array"
        ):
            return False
        return True
    if "number" == tS:
        if validator.is_type(instance, "boolean"):
            return False
        try:
            float(instance)
            return True
        except (ValueError, TypeError):
            return False
    if "integer" == tS:
        if validator.is_type(instance, "boolean"):
            return False
        try:
            int(instance)
            return True
        except (ValueError, TypeError):
            return False
    if "boolean" == tS:
        if validator.is_type(instance, "boolean"):
            return True
        if instance in list(BOOLEAN_STRINGS):
            return True
    return False


def _raw_type_strict(validator: Validator, tS: Any, instance: Any) -> bool:
    return validator.is_type(instance, tS)


# pylint: disable=unused-argument
def cfn_type(validator: Validator, tS: Any, instance: Any, schema: Any):
    """
    When evaluating a type in CloudFormation we have to account
    for the intrinsic functions that the values can represent
    (Ref, GetAtt, If, ...).  This will evaluate if the correct type
    is not found and the instance is an object with a function
    that we do our best to evaluate if that function represents the
    type we are looking for
    """
    if validator.context.strict_types:
        raw_type_fn = _raw_type_strict
    else:
        raw_type_fn = _raw_type
    tS = ensure_list(tS)
    if not any(raw_type_fn(validator, type, instance) for type in tS):
        reprs = ", ".join(repr(type) for type in tS)
        yield ValidationError(f"{instance!r} is not of type {reprs}")


cfn_validators: dict[str, V] = {
    "additionalProperties": additionalProperties,
    "fn_items": FnItems().validate,
    "type": cfn_type,
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_dynamodb_table
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class TableBillingModeExclusive(CfnLintJsonSchema):
    id = "E3638"
    shortdesc = "Validate DynamoDB BillingMode pay per request configuration"
    description = (
        "When 'BillingMode' is 'PAY_PER_REQUEST' don't specify 'ProvisionedThroughput'"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::DynamoDB::Table/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_dynamodb_table,
                filename="billingmode_exclusive.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "Additional properties are not allowed ('ProvisionedThroughput')"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, singular_types


class FindInMap(BaseFn):
    """Check if FindInMap values are correct"""

    id = "E1011"
    shortdesc = "FindInMap validation of configuration"
    description = "Making sure the function is a list of appropriate config"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-findinmap.html"
    tags = ["functions", "findinmap"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::FindInMap", ("array",) + singular_types, resolved_rule="W1034"
        )

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        scalar_schema = {
            "functions": [
                "Fn::FindInMap",
                "Ref",
            ],
            "schema": {
                "type": ["string"],
            },
        }

        schema = {
            "type": "array",
            "minItems": 3,
            "maxItems": 3,
            "fn_items": [
                scalar_schema,
                scalar_schema,
                scalar_schema,
            ],
        }

        if validator.context.transforms.has_language_extensions_transform():
            scalar_schema["functions"] = [
                "Fn::FindInMap",
                "Fn::Join",
                "Fn::Sub",
                "Fn::If",
                "Fn::Select",
                "Fn::Length",
                "Fn::ToJsonString",
                "Ref",
            ]
            schema["maxItems"] = 4
            schema["fn_items"] = [
                scalar_schema,
                scalar_schema,
                scalar_schema,
                {
                    "functions": [],
                    "schema": {
                        "findinmap_parameters": True,
                        "type": ["object"],
                        "properties": {
                            "DefaultValue": {
                                "default_value": True,
                            }
                        },
                        "additionalProperties": False,
                        "required": ["DefaultValue"],
                    },
                },
            ]

        return schema

    def _default_value(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        validator = validator.evolve(
            context=validator.context.evolve(
                functions=["Ref"],
            ),
        )
        yield from validator.descend(
            instance,
            {
                "type": ("array",) + singular_types,
            },
        )

    def fn_findinmap(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:

        if validator.context.transforms.has_language_extensions_transform():
            # we have to use a special validator for this
            # as we don't want DefaultValue: !Ref AWS::NoValue
            # is valid
            mapping_validator = validator.extend(
                validators={
                    "default_value": self._default_value,
                },
            )(validator.schema)
            validator = mapping_validator.evolve(
                context=validator.context.evolve(
                    resources={},
                ),
                cfn=validator.cfn,
                function_filter=validator.function_filter,
                resolver=validator.resolver,
            )
            yield from super().validate(validator, s, instance, schema)
            return

        validator = validator.evolve(
            context=validator.context.evolve(
                resources={},
            )
        )
        yield from super().validate(validator, s, instance, schema)
